#pragma once

#include <dune/istl/preconditioner.hh>

#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/linearalgebra/eigen/PreconConfig.hpp>
#include <amdis/linearalgebra/eigen/SolverConfig.hpp>

namespace AMDiS
{
  template <class IterativeSolver, class Matrix, class VectorX, class VectorB>
  class IterativeRunner
      : public RunnerInterface<Matrix, VectorX, VectorB>
  {
    using SolverCfg = SolverConfig<IterativeSolver>;
    using PreconCfg = PreconConfig<typename IterativeSolver::Preconditioner>;

  public:
    IterativeRunner(std::string const& prefix)
      : solver_{}
    {
      SolverCfg::init(prefix, solver_);
      PreconCfg::init(prefix + "->precon", solver_.preconditioner());
      Parameters::get(prefix + "->reuse pattern", reusePattern_);
    }


    /// Implementes \ref RunnerInterface::init()
    void init(Matrix const& A) override
    {
      if (!reusePattern_ || !initialized_) {
        solver_.analyzePattern(A);
        initialized_ = true;
      }
      solver_.factorize(A);

      test_exit(solver_.info() == Eigen::Success,
        "Error in solver.compute(matrix)");
    }

    /// Implementes \ref RunnerInterface::exit()
    void exit() override {}

    /// Implementes \ref RunnerInterface::solve()
    int solve(Matrix const& A, VectorX& x, VectorB const& b,
              SolverInfo& solverInfo) override
    {
      solver_.setTolerance(solverInfo.relTolerance());
      x = solver_.solveWithGuess(b, x);

      auto r = VectorB(b);
      if (x.norm() != 0)
        r -= A * x;

      solverInfo.setAbsResidual(r.norm());
      solverInfo.setRelResidual(solver_.error());
      solverInfo.setError(solver_.info());
      msg("number of iteration: {}", solver_.iterations());

      return solver_.info() == Eigen::Success ? 0 : int(solver_.info());
    }

  private:
    IterativeSolver solver_;
    bool reusePattern_ = false;
    bool initialized_ = false;
  };

} // end namespace AMDiS
